
from logging import Logger
from logging import getLogger

from untangle import Element

from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import Size
from pyumldiagrams.xmlsupport import XmlConstants
from pyumldiagrams.xmlsupport.SafeConversions import SafeConversions


class ShapeAttributes(SafeConversions):
    def __init__(self):
        super().__init__()
        self.shapeLogger: Logger = getLogger(__name__)

    def _shapeSize(self, graphicElement: Element) -> Size:

        width:  int = self._stringToInteger(graphicElement[XmlConstants.ATTR_WIDTH_V11])
        height: int = self._stringToInteger(graphicElement[XmlConstants.ATTR_HEIGHT_V11])

        size: Size = Size(width=width, height=height)

        return size

    def _shapePosition(self, graphicElement: Element) -> Position:

        x: int = self._stringToInteger(graphicElement[XmlConstants.ATTR_X_V11])
        y: int = self._stringToInteger(graphicElement[XmlConstants.ATTR_Y_V11])

        position: Position = Position(x=x, y=y)

        return position




from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element
from untangle import parse

from codeallyadvanced.ui.AttachmentSide import AttachmentSide

from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import ClassDefinitions
from pyumldiagrams.Definitions import DisplayMethodParameters
from pyumldiagrams.Definitions import FieldDefinition
from pyumldiagrams.Definitions import Fields
from pyumldiagrams.Definitions import MethodDefinition
from pyumldiagrams.Definitions import Methods
from pyumldiagrams.Definitions import NoteDefinition
from pyumldiagrams.Definitions import ParameterDefinition
from pyumldiagrams.Definitions import Parameters
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import UmlLineDefinitions
from pyumldiagrams.Definitions import UmlLollipopDefinition
from pyumldiagrams.Definitions import UmlLollipopDefinitions
from pyumldiagrams.Definitions import UmlNoteDefinitions
from pyumldiagrams.Definitions import VisibilityType
from pyumldiagrams.Definitions import createFieldsFactory
from pyumldiagrams.Definitions import createMethodsFactory
from pyumldiagrams.Definitions import createParametersFactory

from pyumldiagrams.Internal import Elements

from pyumldiagrams.xmlsupport import XmlConstants

from pyumldiagrams.xmlsupport.AbstractToClassDefinition import AbstractToClassDefinition
from pyumldiagrams.xmlsupport.UnTangleLineDefinition import UnTangleLineDefinition
from pyumldiagrams.xmlsupport.UnTangleNoteDefinition import UnTangleNoteDefinition

from pyumldiagrams.xmlsupport.exceptions.MultiDocumentException import MultiDocumentException
from pyumldiagrams.xmlsupport.exceptions.NotClassDiagramException import NotClassDiagramException


class UnTangleToClassDefinition(AbstractToClassDefinition):
    """
    The V11 version uses 'untangle' to parse XML files, as opposed to minidom
    Only untangles a single document at a time
    """

    def __init__(self, fqFileName: str):
        super().__init__()

        self.logger: Logger  = getLogger(__name__)

        self._xmlString: str     = ''
        self._root:     Element = cast(Element, None)
        self._project:  Element = cast(Element, None)
        self._document: Element = cast(Element, None)

        self._umlNoteDefinitions:     UmlNoteDefinitions     = UmlNoteDefinitions([])
        self._umlLollipopDefinitions: UmlLollipopDefinitions = UmlLollipopDefinitions([])

        with open(fqFileName) as xmlFile:
            self._xmlString = xmlFile.read()
            self._root      = parse(self._xmlString)
            self._project   = self._root.PyutProject

            documentElements: Elements = self._project.get_elements(XmlConstants.ELEMENT_DOCUMENT_V11)
            if len(documentElements) != 1:
                raise MultiDocumentException()

        self._document   = documentElements[0]
        if self._document[XmlConstants.PROJECT_TYPE_V11] != 'CLASS_DIAGRAM':
            raise NotClassDiagramException()

        self.logger.debug(f'{self._xmlString=}')
        self.logger.debug(f'{self._root=}')
        self.logger.debug(f'{self._project=}')
        self.logger.debug(f'{self._document=}')

    def generateDefinitions(self):
        """
        Convenience method
        """
        self.generateClassDefinitions()
        self.generateUmlLineDefinitions()
        self.generateUmlNoteDefinitions()
        self.generateLollipopDefinitions()

    def generateClassDefinitions(self):

        pyutDocument = self._document

        graphicElements: Elements = pyutDocument.get_elements(XmlConstants.ELEMENT_GRAPHIC_CLASS_V11)

        for graphicElement in graphicElements:

            self.logger.debug(f'{graphicElement=}')

            pyutClassElement: Element         = graphicElement.PyutClass
            classDefinition:  ClassDefinition = ClassDefinition(name=pyutClassElement[XmlConstants.ATTR_NAME_V11])

            classDefinition.size     = self._shapeSize(graphicElement=graphicElement)
            classDefinition.position = self._shapePosition(graphicElement=graphicElement)

            classDefinition = self._classAttributes(pyutElement=pyutClassElement, classDefinition=classDefinition)

            classDefinition.methods = self._generateMethods(pyutElement=pyutClassElement)
            classDefinition.fields  = self._generateFields(pyutClassElement=pyutClassElement)

            self._classDefinitions.append(classDefinition)

        self.logger.info(f'Generated {len(self.classDefinitions)} class definitions')

    def generateUmlLineDefinitions(self):

        pyutDocument = self._document

        linkElements:           Elements               = pyutDocument.get_elements(XmlConstants.ELEMENT_GRAPHIC_LINK_V11)
        untangleLineDefinition: UnTangleLineDefinition = UnTangleLineDefinition()

        for linkElement in linkElements:
            umlLineDefinition: UmlLineDefinition = untangleLineDefinition.untangle(linkElement=linkElement)
            self._umlLineDefinitions.append(umlLineDefinition)

    def generateLollipopDefinitions(self):
        pyutDocument = self._document

        lollipopElements: Elements = pyutDocument.get_elements(XmlConstants.ELEMENT_GRAPHIC_LOLLIPOP_V11)
        for lollipopElement in lollipopElements:
            umlLollipopDefinition: UmlLollipopDefinition = UmlLollipopDefinition()

            x:              int = self._stringToInteger(lollipopElement[XmlConstants.ATTR_X_V11])
            y:              int = self._stringToInteger(lollipopElement[XmlConstants.ATTR_Y_V11])

            attachmentSide: str = lollipopElement[XmlConstants.ATTR_ATTACHMENT_SIDE_V11]

            pyutLollipopElement: Element = lollipopElement.PyutInterface
            name:                str     = pyutLollipopElement[XmlConstants.ATTR_NAME_V11]

            umlLollipopDefinition.name           = name
            umlLollipopDefinition.position       = Position(x=x, y=y)
            umlLollipopDefinition.attachmentSide = AttachmentSide.toEnum(attachmentSide)

            self._umlLollipopDefinitions.append(umlLollipopDefinition)

    def generateUmlNoteDefinitions(self):
        pyutDocument = self._document

        noteElements: Elements = pyutDocument.get_elements(XmlConstants.ELEMENT_GRAPHIC_NOTE_V11)
        untangleNoteDefinition: UnTangleNoteDefinition = UnTangleNoteDefinition()
        for noteElement in noteElements:
            noteDefinition: NoteDefinition = untangleNoteDefinition.untangle(linkElement=noteElement)

            self._umlNoteDefinitions.append(noteDefinition)

    @property
    def classDefinitions(self) -> ClassDefinitions:
        return self._classDefinitions

    @property
    def umlLineDefinitions(self) -> UmlLineDefinitions:
        return self._umlLineDefinitions

    @property
    def umlNoteDefinitions(self) -> UmlNoteDefinitions:
        return self._umlNoteDefinitions

    @property
    def umlLollipopDefinitions(self) -> UmlLollipopDefinitions:
        return self._umlLollipopDefinitions

    def _generateMethods(self, pyutElement: Element) -> Methods:

        methods: Methods = createMethodsFactory()

        methodElements: Elements = pyutElement.get_elements(XmlConstants.ELEMENT_MODEL_METHOD_V11)
        for methodElement in methodElements:

            methodName: str               = methodElement[XmlConstants.ATTR_NAME_V11]
            method:     MethodDefinition = MethodDefinition(name=methodName)

            visibilityStr: str = methodElement[XmlConstants.ATTR_VISIBILITY_V11]
            method.visibility  = VisibilityType.toEnum(visibilityStr)
            method.returnType  = methodElement[XmlConstants.ATTR_RETURN_TYPE_V11]

            method.parameters = self._generateParameters(methodElement=methodElement)

            methods.append(method)

        return methods

    def _generateParameters(self, methodElement: Element) -> Parameters:

        parameters: Parameters = createParametersFactory()

        parameterElements: Elements = methodElement.get_elements(XmlConstants.ELEMENT_MODEL_PARAMETER_V11)
        for parameterElement in parameterElements:

            parameterName:       str                 = parameterElement[XmlConstants.ATTR_NAME_V11]
            parameterDefinition: ParameterDefinition = ParameterDefinition(name=parameterName)

            parameterDefinition.defaultValue  = parameterElement[XmlConstants.ATTR_DEFAULT_VALUE_V11]
            parameterDefinition.parameterType = parameterElement[XmlConstants.ATTR_TYPE_V11]
            parameters.append(parameterDefinition)

        return parameters

    def _generateFields(self, pyutClassElement: Element) -> Fields:

        fields: Fields = createFieldsFactory()

        fieldElements: Elements = pyutClassElement.get_elements(XmlConstants.ELEMENT_MODEL_FIELD_V11)
        for fieldElement in fieldElements:

            fieldName:       str             = fieldElement[XmlConstants.ATTR_NAME_V11]
            fieldDefinition: FieldDefinition = FieldDefinition(name=fieldName)

            visibilityStr: str = fieldElement[XmlConstants.ATTR_VISIBILITY_V11]
            fieldDefinition.visibility    = VisibilityType.toEnum(visibilityStr)

            fieldDefinition.fieldType    = fieldElement[XmlConstants.ATTR_TYPE_V11]
            fieldDefinition.defaultValue = fieldElement[XmlConstants.ATTR_DEFAULT_VALUE_V11]

            fields.append(fieldDefinition)
        return fields

    def _classAttributes(self, pyutElement: Element, classDefinition: ClassDefinition) -> ClassDefinition:

        classDefinition.displayMethods          = self._stringToBoolean(pyutElement[XmlConstants.ATTR_DISPLAY_METHODS_V11])
        classDefinition.displayMethodParameters = DisplayMethodParameters(pyutElement[XmlConstants.ATTR_DISPLAY_PARAMETERS_V11])
        classDefinition.displayFields           = self._stringToBoolean(pyutElement[XmlConstants.ATTR_DISPLAY_FIELDS_V11])
        classDefinition.displayStereotype       = self._stringToBoolean(pyutElement[XmlConstants.ATTR_DISPLAY_STEREOTYPE_V11])
        classDefinition.fileName                = pyutElement[XmlConstants.ATTR_FILENAME_V11]
        classDefinition.description             = pyutElement[XmlConstants.ATTR_DESCRIPTION_V11]

        return classDefinition




from logging import Logger
from logging import getLogger

from abc import abstractmethod
from abc import ABCMeta

from pyumldiagrams.Definitions import ClassDefinitions
from pyumldiagrams.Definitions import UmlLineDefinitions

from pyumldiagrams.xmlsupport.ShapeAttributes import ShapeAttributes


class MyMetaAbstractToClassDefinition(ABCMeta, type(ShapeAttributes)):        # type: ignore
    """
    I have know idea why this works:
    https://stackoverflow.com/questions/66591752/metaclass-conflict-when-trying-to-create-a-python-abstract-class-that-also-subcl
    """
    pass


class AbstractToClassDefinition(ShapeAttributes):
    __metaclass__ = MyMetaAbstractToClassDefinition

    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._classDefinitions:   ClassDefinitions   = ClassDefinitions([])
        self._umlLineDefinitions: UmlLineDefinitions = UmlLineDefinitions([])

    @abstractmethod
    def generateClassDefinitions(self):
        pass

    @abstractmethod
    def generateUmlLineDefinitions(self):
        pass

    @property
    @abstractmethod
    def classDefinitions(self) -> ClassDefinitions:
        pass

    @property
    @abstractmethod
    def umlLineDefinitions(self) -> UmlLineDefinitions:
        pass




ELEMENT_DOCUMENT:      str = 'PyutDocument'
TOP_LEVEL_ELEMENT:     str = 'PyutProject'
ELEMENT_GRAPHIC_CLASS: str = 'GraphicClass'
ELEMENT_GRAPHIC_LINK:  str = 'GraphicLink'
ELEMENT_MODEL_CLASS:   str = 'Class'
ELEMENT_MODEL_LINK:    str = 'Link'
ELEMENT_CONTROL_POINT: str = 'ControlPoint'

ELEMENT_MODEL_METHOD: str = 'Method'

ATTR_WIDTH:  str = 'width'
ATTR_HEIGHT: str = 'height'
ATTR_X:      str = 'x'
ATTR_Y:      str = 'y'
ATTR_NAME:   str = 'name'

ATTR_LINK_SOURCE_ANCHOR_X: str = 'srcX'
ATTR_LINK_SOURCE_ANCHOR_Y: str = 'srcY'

ATTR_LINK_DESTINATION_ANCHOR_X: str = 'dstX'
ATTR_LINK_DESTINATION_ANCHOR_Y: str = 'dstY'

ATTR_TYPE: str = 'type'

ATTR_SHOW_METHODS:      str = 'showMethods'
ATTR_SHOW_FIELDS:       str = 'showFields'
ATTR_SHOW_STEREOTYPE:   str = 'showStereotype'

ATTR_DISPLAY_PARAMETERS: str = 'displayParameters'

ELEMENT_MODEL_PARAM:       str = 'Param'
ELEMENT_MODEL_RETURN_TYPE: str = 'Return'
ATTR_DEFAULT_VALUE:  str = 'defaultValue'
ATTR_VISIBILITY:     str = 'visibility'

ELEMENT_PROJECT_V11:          str = 'PyutProject'
PROJECT_TYPE_V11:             str = 'type'
ELEMENT_DOCUMENT_V11:         str = 'PyutDocument'
ELEMENT_GRAPHIC_CLASS_V11:    str = 'OglClass'
ELEMENT_GRAPHIC_LINK_V11:     str = 'OglLink'
ELEMENT_CONTROL_POINT_V11:    str = 'ControlPoint'
ELEMENT_GRAPHIC_NOTE_V11:     str = 'OglNote'
ELEMENT_GRAPHIC_LOLLIPOP_V11: str = 'OglInterface2'

ELEMENT_LABEL_CENTER_V11:      str = 'LabelCenter'
ELEMENT_LABEL_SOURCE_V11:      str = 'LabelSource'
ELEMENT_LABEL_DESTINATION_V11: str = 'LabelDestination'

ATTR_WIDTH_V11:  str = 'width'
ATTR_HEIGHT_V11: str = 'height'
ATTR_X_V11:      str = 'x'
ATTR_Y_V11:      str = 'y'
ATTR_NAME_V11:   str = 'name'


ATTR_STEREOTYPE_V11:            str = 'stereotype'
ATTR_DISPLAY_METHODS_V11:       str = 'displayMethods'
ATTR_DISPLAY_PARAMETERS_V11:    str = 'displayParameters'
ATTR_DISPLAY_FIELDS_V11:        str = 'displayFields'
ATTR_DISPLAY_STEREOTYPE_V11:    str = 'displayStereotype'
ATTR_FILENAME_V11:              str = 'fileName'
ATTR_DESCRIPTION_V11:           str = 'description'

ELEMENT_MODEL_METHOD_V11: str = 'PyutMethod'
ATTR_VISIBILITY_V11:      str = 'visibility'
ATTR_RETURN_TYPE_V11:     str = 'returnType'

ELEMENT_MODEL_PARAMETER_V11: str  = 'PyutParameter'
ATTR_DEFAULT_VALUE_V11:      str  = 'defaultValue'
ATTR_TYPE_V11:               str  = 'type'

ELEMENT_MODEL_FIELD_V11:    str = 'PyutField'

ATTR_LINK_SOURCE_ANCHOR_X_V11: str = 'sourceAnchorX'
ATTR_LINK_SOURCE_ANCHOR_Y_V11: str = 'sourceAnchorY'

ATTR_LINK_DESTINATION_ANCHOR_X_V11: str = 'destinationAnchorX'
ATTR_LINK_DESTINATION_ANCHOR_Y_V11: str = 'destinationAnchorY'

ATTR_CARDINALITY_SOURCE_V11:      str = 'cardinalitySource'
ATTR_CARDINALITY_DESTINATION_V11: str = 'cardinalityDestination'

ATTR_CONTENT_V11:         str = 'content'
ATTR_ATTACHMENT_SIDE_V11: str = 'attachmentPoint'




from logging import Logger
from logging import getLogger
from typing import cast

from untangle import Element

from pyumldiagrams.Definitions import LinePositions
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import NamedAssociations
from pyumldiagrams.Internal import Elements

from pyumldiagrams.xmlsupport import XmlConstants

from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.xmlsupport.SafeConversions import SafeConversions


class UnTangleLineDefinition(SafeConversions):
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

    def untangle(self, linkElement: Element) -> UmlLineDefinition:

        linePositions: LinePositions = self._getLinePositions(linkElement=linkElement)

        pyutLinkElement: Element  = linkElement.PyutLink
        strType:         str      = pyutLinkElement[XmlConstants.ATTR_TYPE_V11]
        lineType:        LineType = LineType.toEnum(strType)

        umlLineDefinition: UmlLineDefinition = UmlLineDefinition(linePositions=linePositions, lineType=lineType)

        if lineType in NamedAssociations:

            nameStr: str = pyutLinkElement[XmlConstants.ATTR_NAME_V11]
            umlLineDefinition.name = nameStr

            umlLineDefinition.cardinalitySource      = pyutLinkElement[XmlConstants.ATTR_CARDINALITY_SOURCE_V11]
            umlLineDefinition.cardinalityDestination = pyutLinkElement[XmlConstants.ATTR_CARDINALITY_DESTINATION_V11]

            umlLineDefinition.namePosition                   = self._getNamePosition(linkElement=linkElement)
            umlLineDefinition.sourceCardinalityPosition      = self._getSourcePosition(linkElement=linkElement)
            umlLineDefinition.destinationCardinalityPosition = self._getDestinationPosition(linkElement=linkElement)

        elif lineType == LineType.NoteLink or lineType == LineType.Interface:
            nameStr = pyutLinkElement[XmlConstants.ATTR_NAME_V11]
            umlLineDefinition.name = nameStr

        return umlLineDefinition

    def _getLinePositions(self, linkElement: Element) -> LinePositions:

        srcPosition: Position = self._getStartPosition(linkElement=linkElement)

        linePositions:     LinePositions     = LinePositions([srcPosition])

        # Intermediate points
        controlPoints: Elements = linkElement.get_elements(XmlConstants.ELEMENT_CONTROL_POINT_V11)
        for point in controlPoints:
            controlPointElement: Element = cast(Element, point)

            self.logger.debug(f'{controlPointElement=}')
            x: int = self._stringToInteger(controlPointElement[XmlConstants.ATTR_X_V11])
            y: int = self._stringToInteger(controlPointElement[XmlConstants.ATTR_Y_V11])
            bendPosition: Position = Position(x=x, y=y)
            linePositions.append(bendPosition)

        # end of line
        destX: int = self._stringToInteger(linkElement[XmlConstants.ATTR_LINK_DESTINATION_ANCHOR_X_V11])
        destY: int = self._stringToInteger(linkElement[XmlConstants.ATTR_LINK_DESTINATION_ANCHOR_Y_V11])

        destPosition: Position = Position(x=destX, y=destY)

        linePositions.append(destPosition)

        return linePositions

    def _getStartPosition(self, linkElement: Element) -> Position:

        srcX: int = self._stringToInteger(linkElement[XmlConstants.ATTR_LINK_SOURCE_ANCHOR_X_V11])
        srcY: int = self._stringToInteger(linkElement[XmlConstants.ATTR_LINK_SOURCE_ANCHOR_Y_V11])

        srcPosition: Position = Position(x=srcX, y=srcY)

        return srcPosition

    def _getNamePosition(self, linkElement: Element) -> Position:
        return self._getLabelPosition(linkElement=linkElement, elementName=XmlConstants.ELEMENT_LABEL_CENTER_V11)

    def _getSourcePosition(self, linkElement: Element) -> Position:
        return self._getLabelPosition(linkElement=linkElement, elementName=XmlConstants.ELEMENT_LABEL_SOURCE_V11)

    def _getDestinationPosition(self, linkElement: Element) -> Position:
        return self._getLabelPosition(linkElement=linkElement, elementName=XmlConstants.ELEMENT_LABEL_DESTINATION_V11)

    def _getLabelPosition(self, linkElement: Element, elementName: str) -> Position:

        positionElements: Elements = linkElement.get_elements(elementName)

        assert len(positionElements) == 1, f'There can only be one ....  position element -- {elementName}'

        positionElement: Element = positionElements[0]

        x: int = self._stringToInteger(positionElement[XmlConstants.ATTR_X_V11])
        y: int = self._stringToInteger(positionElement[XmlConstants.ATTR_Y_V11])

        position: Position = Position(x=x, y=y)

        return position




from logging import Logger
from logging import getLogger


class SafeConversions:
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

    def _stringToBoolean(self, strBoolValue: str) -> bool:

        self.logger.debug(f'{strBoolValue=}')
        try:
            if strBoolValue is not None:
                if strBoolValue in [True, "True", "true", 1, "1"]:
                    return True
        except (ValueError, Exception) as e:
            self.logger.error(f'_stringToBoolean error: {e}')

        return False

    def _stringToInteger(self, x: str):
        if x is not None and x != '':
            return int(x)
        else:
            return 0




from typing import Final
from typing import List
from typing import NewType
from typing import Tuple

from logging import Logger
from logging import getLogger
from typing import cast

from fpdf import FPDF

from pyumldiagrams.Common import Common

from pyumldiagrams.CommonAbsolute import CommonAbsolute

from pyumldiagrams.IDiagramLine import IDiagramLine

from pyumldiagrams.UnsupportedException import UnsupportedException

from pyumldiagrams.Internal import ArrowPoints
from pyumldiagrams.Internal import DiamondPoints
from pyumldiagrams.Internal import PolygonPoints
from pyumldiagrams.Internal import InternalPosition

from pyumldiagrams.Definitions import LinePositions
from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import Position

from pyumldiagrams.Definitions import UmlLollipopDefinition


from pyumldiagrams.pdf.PdfCommon import PdfCommon
from pyumldiagrams.pdf.PdfLollipopInterface import PdfLollipopInterface


PDFCoordinate  = NewType('PDFCoordinate',  Tuple[int, int])
PDFCoordinates = NewType('PDFCoordinates', List[PDFCoordinate])


PolyPoint  = NewType('PolyPoint',  Tuple[int, int])
PolyPoints = NewType('PolyPoints', List[PolyPoint])


class PdfLine(IDiagramLine):
    """
    This class takes responsibility for drawing the various types of lines within the
    described UML classes.  End users generally do not directly use this class.
    It is split off as part of the separation of responsibility principle.
    """
    INHERITANCE_ARROW_HEIGHT: Final = 10
    DIAMOND_HEIGHT:           Final = 8

    DASH_LENGTH:     Final = 5.0
    DASH_GAP_LENGTH: Final = 7.0

    def __init__(self, pdf: FPDF, diagramPadding: DiagramPadding, dpi: int):

        super().__init__(docMaker=pdf, diagramPadding=diagramPadding, dpi=dpi)
        self.logger: Logger = getLogger(__name__)

        self._pdfLollipopInterface: PdfLollipopInterface = PdfLollipopInterface(pdf=pdf, dpi=dpi)
        self._pdfCommon:            PdfCommon            = PdfCommon(diagramPadding=diagramPadding,  dpi=dpi)

    def draw(self, lineDefinition: UmlLineDefinition):
        """
        Draw the line described by the input parameter
        Args:
            lineDefinition:  Describes the line to draw
        """

        linePositions: LinePositions = lineDefinition.linePositions
        lineType:      LineType      = lineDefinition.lineType

        if lineType == LineType.Inheritance:
            self._drawInheritance(linePositions=linePositions)
        elif lineType == LineType.Composition:
            self._drawCompositeAggregation(lineDefinition=lineDefinition)
        elif lineType == LineType.Aggregation:
            self._drawSharedAggregation(lineDefinition=lineDefinition)
        elif lineType == LineType.Association:
            self._drawAssociation(lineDefinition=lineDefinition)
        elif lineType == LineType.Interface:
            self._drawInterface(lineDefinition=lineDefinition)
        elif lineType == LineType.NoteLink:
            self._drawNoteLink(lineDefinition=lineDefinition)
        else:
            raise UnsupportedException(f'Line definition type not supported: `{lineType}`')

    def drawLollipopInterface(self, umlLollipopDefinition: UmlLollipopDefinition):

        self._pdfLollipopInterface.drawLollipopInterface(umlLollipopDefinition=umlLollipopDefinition)

    def _drawInheritance(self, linePositions: LinePositions):
        """
        Must account for the margins and gaps between drawn shapes
        Must convert to from screen noteCoordinates to point noteCoordinates
        Draw the arrow first
        Compute the mid-point of the bottom line of the arrow
        That is where the line ends

        Args:
            linePositions - The points that describe the line
        """
        points:  ArrowPoints = self._drawArrowAtDestination(linePositions)

        self._drawLineSourceToDestination(linePositions, points)

    def _drawInterface(self, lineDefinition: UmlLineDefinition):

        points:  ArrowPoints = self._drawArrowAtDestination(lineDefinition.linePositions)

        self._drawLineSourceToDestination(lineDefinition.linePositions, points, dashedLine=True)

    def _drawCompositeAggregation(self, lineDefinition: UmlLineDefinition):
        """
        Composition

        Args:
            lineDefinition:
        """
        self._drawAggregation(lineDefinition=lineDefinition, isComposite=True)

    def _drawSharedAggregation(self, lineDefinition: UmlLineDefinition):
        """
        Aggregation

        Args:
            lineDefinition:
        """
        self._drawAggregation(lineDefinition=lineDefinition, isComposite=False)

    def _drawAssociation(self, lineDefinition: UmlLineDefinition):

        pdf:           FPDF          = self._docMaker
        linePositions: LinePositions = lineDefinition.linePositions

        polyPoints: PolyPoints = self._toPolyPoints(linePositions=linePositions)

        pdf.polyline(polyPoints, style='D')

        self._drawAssociationName(lineDefinition=lineDefinition)
        self._drawSourceCardinality(lineDefinition=lineDefinition)
        self._drawDestinationCardinality(lineDefinition=lineDefinition)

    def _drawNoteLink(self, lineDefinition: UmlLineDefinition):

        pdf:           FPDF          = self._docMaker
        linePositions: LinePositions = lineDefinition.linePositions
        polyPoints:    PolyPoints = self._toPolyPoints(linePositions=linePositions)

        with pdf.local_context():
            pdf.set_dash_pattern(dash=PdfLine.DASH_LENGTH, gap=PdfLine.DASH_GAP_LENGTH)
            pdf.polyline(polyPoints, style='D')

    def _drawAggregation(self, lineDefinition: UmlLineDefinition, isComposite: bool):

        linePositions: LinePositions = lineDefinition.linePositions

        startPoints: Tuple[InternalPosition, InternalPosition] = self._convertPoints(linePositions[0], linePositions[1])

        convertedSrc: InternalPosition = startPoints[0]
        convertedDst: InternalPosition = startPoints[1]

        points: DiamondPoints = Common.computeDiamondVertices(position0=convertedSrc, position1=convertedDst)
        self._drawDiamond(diamondPoints=points, isComposite=isComposite)

        newEndStartPoint: InternalPosition = points[3]

        self._finishDrawingLine(linePositions=linePositions, newStartPoint=newEndStartPoint)

        self._drawAssociationName(lineDefinition=lineDefinition)
        self._drawSourceCardinality(lineDefinition=lineDefinition)
        self._drawDestinationCardinality(lineDefinition=lineDefinition)

    def _drawDiamond(self, diamondPoints: DiamondPoints, isComposite: bool):

        pdfCoordinates: PDFCoordinates = self._toPDFCoordinates(polygonPoints=diamondPoints)
        pdf:            FPDF           = self._docMaker

        if isComposite is True:
            pdf.polygon(pdfCoordinates, style='DF')
        else:
            pdf.polygon(pdfCoordinates, style='D')

    def _drawAssociationName(self, lineDefinition: UmlLineDefinition):

        pdf: FPDF = self._docMaker

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.namePosition)
        pdf.text(x=iPos.x, y=iPos.y, txt=lineDefinition.name)

    def _drawSourceCardinality(self, lineDefinition: UmlLineDefinition):

        pdf: FPDF = self._docMaker

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.sourceCardinalityPosition)

        pdf.text(x=iPos.x, y=iPos.y, txt=lineDefinition.cardinalitySource)

    def _drawDestinationCardinality(self, lineDefinition: UmlLineDefinition):

        pdf: FPDF = self._docMaker

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.destinationCardinalityPosition)

        pdf.text(x=iPos.x, y=iPos.y, txt=lineDefinition.cardinalityDestination)

    def _drawLineSourceToDestination(self, linePositions: LinePositions, points: ArrowPoints, dashedLine: bool = False):

        pdf: FPDF = self._docMaker

        newEndPosition:    InternalPosition = Common.computeMidPointOfBottomLine(points[0], points[2])
        adjustedPositions: LinePositions    = LinePositions(linePositions[:-1])

        with pdf.local_context():
            if dashedLine is True:
                pdf.set_dash_pattern(dash=PdfLine.DASH_LENGTH, gap=PdfLine.DASH_GAP_LENGTH)

            polyPoints:   PolyPoints = self._toPolyPoints(linePositions=adjustedPositions)
            endPolyPoint: PolyPoint  = PolyPoint((newEndPosition.x, newEndPosition.y))

            polyPoints.append(endPolyPoint)

            pdf.polyline(polyPoints, style='D')

    def _drawArrowAtDestination(self, linePositions: LinePositions) -> ArrowPoints:
        """

        Args:
            linePositions:  The line positions

        Returns:    The points used to draw the arrow
        """

        internalPosition0: InternalPosition = self._pdfCommon.toInternal(linePositions[-1])  # last position
        internalPosition1: InternalPosition = self._pdfCommon.toInternal(linePositions[-2])  # next to last position

        points:         ArrowPoints    = Common.computeTheArrowVertices(position0=internalPosition0, position1=internalPosition1)
        pdfCoordinates: PDFCoordinates = self._toPDFCoordinates(polygonPoints=points)

        self._docMaker.polygon(pdfCoordinates, style='D')

        return points

    def _computeTextPosition(self, lineDefinition: UmlLineDefinition, labelPosition: Position) -> InternalPosition:

        xy: Tuple[int, int] = CommonAbsolute.computeAbsoluteLabelPosition(srcPosition=lineDefinition.linePositions[0],
                                                                          dstPosition=lineDefinition.linePositions[-1],
                                                                          labelPosition=labelPosition)

        iPos: InternalPosition = self._pdfCommon.toInternal(position=Position(x=xy[0], y=xy[1]))

        return iPos

    def _toPDFCoordinates(self, polygonPoints: PolygonPoints) -> PDFCoordinates:

        pdfCoordinates: PDFCoordinates = PDFCoordinates([])

        for dPoint in polygonPoints:
            diamondPoint: InternalPosition = cast(InternalPosition, dPoint)
            pdfTuple:     PDFCoordinate    = PDFCoordinate((diamondPoint.x, diamondPoint.y))
            pdfCoordinates.append(pdfTuple)

        return pdfCoordinates

    def _finishDrawingLine(self, linePositions: LinePositions, newStartPoint: InternalPosition):
        """
        Finishes drawing the line for aggregation/composition where the diamond is at the source;  Thus,
        the new start position at a diamond tip

        Args:
            linePositions:
            newStartPoint:
        """

        linePositionsCopy: LinePositions = LinePositions(linePositions[1:])    # Makes a copy; remove first one
        pdf:               FPDF          = self._docMaker

        currentPos:          Position         = linePositionsCopy[0]
        convertedCurrentPos: InternalPosition = self._pdfCommon.toInternal(position=currentPos)
        pdf.line(x1=newStartPoint.x, y1=newStartPoint.y, x2=convertedCurrentPos.x, y2=convertedCurrentPos.y)

        polyPoints: PolyPoints = self._toPolyPoints(linePositions=linePositionsCopy)
        pdf.polyline(polyPoints, style='D')

    def _convertPoints(self, src: Position, dst: Position) -> Tuple[InternalPosition, InternalPosition]:

        convertedSrc: InternalPosition = self._pdfCommon.toInternal(position=src)
        convertedDst: InternalPosition = self._pdfCommon.toInternal(position=dst)

        return convertedSrc, convertedDst

    def _toPolyPoints(self, linePositions: LinePositions) -> PolyPoints:
        """
        Takes the raw positions and creates the poly points associated with our adjusted internal positions.
        Args:
            linePositions:

        Returns:  PDF style internal points
        """

        polyPoints: PolyPoints = PolyPoints([])

        for position in linePositions:
            lastInternal: InternalPosition = self._pdfCommon.toInternal(position)

            polyPoint: PolyPoint = PolyPoint((lastInternal.x, lastInternal.y))
            polyPoints.append(polyPoint)

        return polyPoints




from typing import Any
from typing import List
from typing import NewType
from typing import Tuple
from typing import Union
from typing import Final

from logging import Logger
from logging import getLogger

from os import sep as osSep

from PIL.ImageDraw import ImageDraw
from PIL.ImageFont import ImageFont
from PIL.ImageFont import truetype

from codeallybasic.ResourceManager import ResourceManager

from pyumldiagrams.BaseDiagram import BaseDiagram
from pyumldiagrams.Common import Common
from pyumldiagrams.CommonAbsolute import CommonAbsolute
from pyumldiagrams.IDiagramLine import IDiagramLine
from pyumldiagrams.UnsupportedException import UnsupportedException

from pyumldiagrams.image.ImageCommon import ImageCommon

from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import LinePositions

from pyumldiagrams.Internal import ArrowPoints
from pyumldiagrams.Internal import DiamondPoints
from pyumldiagrams.Internal import InternalPosition

PILPoints     = NewType('PILPoints',     List[int])         # Maybe these go
PolygonPoints = NewType('PolygonPoints', List[int])         # in Internal


class ImageLine(IDiagramLine):

    DEFAULT_LINE_COLOR: Final = 'Black'
    DEFAULT_TEXT_COLOR: Final = 'Black'

    LINE_WIDTH:         Final = 1

    RESOURCES_PACKAGE_NAME: Final = 'pyumldiagrams.image.resources'
    RESOURCES_PATH:         Final = f'pyumldiagrams{osSep}image{osSep}resources'

    def __init__(self, docWriter: Any, diagramPadding: DiagramPadding):

        super().__init__(docMaker=docWriter, diagramPadding=diagramPadding, dpi=0)

        self.logger: Logger = getLogger(__name__)

        self._imgDraw: ImageDraw = docWriter

        # noinspection SpellCheckingInspection
        fqPath:     str       = self._retrieveResourcePath('MonoFonto.ttf')
        self._font: ImageFont = truetype(font=fqPath, size=BaseDiagram.DEFAULT_FONT_SIZE)

    def draw(self, lineDefinition: UmlLineDefinition):
        """
        Draw the line described by the input parameter

        Args:
            lineDefinition:  Describes the line to draw
        """
        linePositions: LinePositions = lineDefinition.linePositions
        lineType:      LineType      = lineDefinition.lineType

        if lineType == LineType.Inheritance:
            self._drawInheritance(linePositions=linePositions)
        elif lineType == LineType.Composition:
            self._drawCompositeAggregation(lineDefinition=lineDefinition)
        elif lineType == LineType.Aggregation:
            self._drawSharedAggregation(lineDefinition=lineDefinition)
        elif lineType == LineType.Association:
            self._drawAssociation(lineDefinition=lineDefinition)
        elif lineType == LineType.Interface:
            pass        # TODO
        elif lineType == LineType.NoteLink:
            pass        # TODO
        else:
            raise UnsupportedException(f'Line definition type not supported: `{lineType}`')

    def _drawInheritance(self, linePositions: LinePositions):
        """
        Must account for the margins and gaps between drawn shapes
        Must convert from screen coordinates to point coordinates
        Draw the arrow first
        Compute the mid-point of the bottom line of the arrow
        That is where the line ends

        Args:
            linePositions  The points that describe the line
        """
        internalPosition0:  InternalPosition = self._toInternal(linePositions[-1])
        internalPosition1:  InternalPosition = self._toInternal(linePositions[-2])

        points:  ArrowPoints   = Common.computeTheArrowVertices(position0=internalPosition0, position1=internalPosition1)
        polygon: PolygonPoints = self._toPolygonPoints(points)

        self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR)

        newEndPoint: InternalPosition = Common.computeMidPointOfBottomLine(points[0], points[2])

        xy: PILPoints = PILPoints([])

        adjustedPositions: LinePositions = LinePositions(linePositions[:-1])
        for externalPosition in adjustedPositions:
            internalPosition: InternalPosition = self._toInternal(externalPosition)
            xy.append(internalPosition.x)
            xy.append(internalPosition.y)
        xy.append(newEndPoint.x)
        xy.append(newEndPoint.y)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

    def _drawCompositeAggregation(self, lineDefinition: UmlLineDefinition):
        """
        Composition
        Draws both the line and the solid diamond

        Args:
            lineDefinition:   The line definition
        """
        self._drawAggregation(lineDefinition=lineDefinition, isComposite=True)

    def _drawSharedAggregation(self, lineDefinition: UmlLineDefinition):
        """
        Aggregation
        Draws both the line and the hollow diamond

        Args:
            lineDefinition:   The line definition
        """
        self._drawAggregation(lineDefinition=lineDefinition, isComposite=False)

    def _drawAggregation(self, lineDefinition: UmlLineDefinition, isComposite: bool):
        """

        Args:
            lineDefinition: The UML line definitions
            isComposite:    'True' draws solid composition diamond;  'False' draws hollow aggregation diamond
        """
        linePositions: LinePositions = lineDefinition.linePositions

        internalPosition0:  InternalPosition = self._toInternal(linePositions[0])
        internalPosition1:  InternalPosition = self._toInternal(linePositions[1])

        points:  DiamondPoints = Common.computeDiamondVertices(position1=internalPosition1, position0=internalPosition0)
        polygon: PolygonPoints = self._toPolygonPoints(points)

        if isComposite is True:
            self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR, fill='black')
        else:
            self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR)

        newStartPoint: InternalPosition = points[3]
        xy:            PILPoints        = self._toPILPoints(linePositions=linePositions, newStartPoint=newStartPoint)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

        self._drawAssociationName(lineDefinition=lineDefinition)
        self._drawSourceCardinality(lineDefinition=lineDefinition)
        self._drawDestinationCardinality(lineDefinition=lineDefinition)

    def _drawAssociation(self, lineDefinition: UmlLineDefinition):

        linePositions: LinePositions = lineDefinition.linePositions
        xy:            PILPoints     = PILPoints([])

        for externalPosition in linePositions:
            internalPosition: InternalPosition = self._toInternal(externalPosition)
            xy.append(internalPosition.x)
            xy.append(internalPosition.y)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

        self._drawAssociationName(lineDefinition=lineDefinition)
        self._drawSourceCardinality(lineDefinition=lineDefinition)
        self._drawDestinationCardinality(lineDefinition=lineDefinition)

    def _drawAssociationName(self, lineDefinition: UmlLineDefinition):

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.namePosition)

        self._imgDraw.text(xy=(iPos.x, iPos.y), fill=ImageLine.DEFAULT_TEXT_COLOR, font=self._font, text=lineDefinition.name)

    def _drawSourceCardinality(self, lineDefinition: UmlLineDefinition):

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.sourceCardinalityPosition)

        self._imgDraw.text(xy=(iPos.x, iPos.y), fill=ImageLine.DEFAULT_TEXT_COLOR, font=self._font, text=lineDefinition.cardinalitySource)

    def _drawDestinationCardinality(self, lineDefinition: UmlLineDefinition):

        iPos: InternalPosition = self._computeTextPosition(lineDefinition=lineDefinition, labelPosition=lineDefinition.destinationCardinalityPosition)

        self._imgDraw.text(xy=(iPos.x, iPos.y), fill=ImageLine.DEFAULT_TEXT_COLOR, font=self._font, text=lineDefinition.cardinalityDestination)

    def _computeTextPosition(self, lineDefinition: UmlLineDefinition, labelPosition: Position) -> InternalPosition:

        xy: Tuple[int, int] = CommonAbsolute.computeAbsoluteLabelPosition(srcPosition=lineDefinition.linePositions[0],
                                                                          dstPosition=lineDefinition.linePositions[-1],
                                                                          labelPosition=labelPosition)

        iPos: InternalPosition = self._toInternal(position=Position(x=xy[0], y=xy[1]))

        return iPos

    def _retrieveResourcePath(self, bareFileName: str) -> str:

        fqFileName: str = ResourceManager.retrieveResourcePath(bareFileName=bareFileName,
                                                               resourcePath=ImageLine.RESOURCES_PATH,
                                                               packageName=ImageLine.RESOURCES_PACKAGE_NAME)

        return fqFileName

    def _toInternal(self, position: Position) -> InternalPosition:

        verticalGap:   int = self._diagramPadding.verticalGap
        horizontalGap: int = self._diagramPadding.horizontalGap

        iPos: InternalPosition = ImageCommon.toInternal(position, verticalGap=verticalGap, horizontalGap=horizontalGap)

        return iPos

    def _toPolygonPoints(self, points: Union[ArrowPoints, DiamondPoints]) -> PolygonPoints:

        polygon: PolygonPoints = PolygonPoints([])

        for point in points:
            polygon.append(int(point.x))
            polygon.append(int(point.y))

        return polygon

    def _toPILPoints(self, linePositions: LinePositions, newStartPoint: InternalPosition) -> PILPoints:

        linePositionsCopy: LinePositions = LinePositions(linePositions[1:])  # Makes a copy; remove old start point

        xy: PILPoints = PILPoints([])

        xy.append(newStartPoint.x)
        xy.append(newStartPoint.y)
        for externalPosition in linePositionsCopy:
            internalPosition: InternalPosition = self._toInternal(externalPosition)
            xy.append(internalPosition.x)
            xy.append(internalPosition.y)

        return xy




from typing import Final
from typing import Tuple

from logging import Logger
from logging import getLogger

from fpdf import FPDF

from codeallybasic.Position import Position

from codeallyadvanced.ui.AttachmentSide import AttachmentSide

from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import RenderStyle
from pyumldiagrams.Definitions import UmlLollipopDefinition

from pyumldiagrams.Internal import InternalPosition
from pyumldiagrams.pdf.PdfCommon import PdfCommon

LOLLIPOP_LINE_LENGTH:   Final = 60
LOLLIPOP_CIRCLE_HEIGHT: Final = 8
LOLLIPOP_CIRCLE_WIDTH:  Final = 8

LOLLIPOP_WIDTH_SQUARED: int = (LOLLIPOP_CIRCLE_WIDTH * LOLLIPOP_CIRCLE_WIDTH)
LOLLIPOP_CIRCLE_RADIUS: int = (LOLLIPOP_CIRCLE_HEIGHT // 2) + (LOLLIPOP_WIDTH_SQUARED // (8 * LOLLIPOP_CIRCLE_HEIGHT))

ADJUST_AWAY_FROM_IMPLEMENTOR: Final = 10


class PdfLollipopInterface:

    def __init__(self, pdf: FPDF, dpi: int):

        self.logger: Logger = getLogger(__name__)

        self._pdf: FPDF = pdf
        self._pdfCommon: PdfCommon = PdfCommon(diagramPadding=DiagramPadding(), dpi=dpi)

    def drawLollipopInterface(self, umlLollipopDefinition: UmlLollipopDefinition):

        position: Position = umlLollipopDefinition.position
        xSrc:     int      = position.x
        ySrc:     int      = position.y

        attachmentSide: AttachmentSide = umlLollipopDefinition.attachmentSide

        circleX, circleY, xDest, yDest = self._calculateWhereToDrawLollipop(attachmentSide, xSrc, ySrc)

        self._createTheInterfaceLine(position=position, xDest=xDest, yDest=yDest)
        self._drawTheTootsiePopCircle(attachmentSide=attachmentSide, circleX=circleX, circleY=circleY)
        self._drawInterfaceName(interfaceName=umlLollipopDefinition.name, attachmentSide=attachmentSide, xSrc=xSrc, ySrc=ySrc)

    def _calculateWhereToDrawLollipop(self, attachmentSide: AttachmentSide, xSrc, ySrc):
        """

        Args:
            attachmentSide:
            xSrc:
            ySrc:

        Returns:  A tuple that is the x,y position of the circle and the end
        of the line
        """

        self.logger.debug(f'({xSrc},{ySrc}) {LOLLIPOP_LINE_LENGTH=}')

        if attachmentSide == AttachmentSide.EAST:
            xDest:   int = int(xSrc + LOLLIPOP_LINE_LENGTH)
            yDest:   int = int(ySrc)
            circleX: int = int(xSrc + LOLLIPOP_LINE_LENGTH)
            circleY: int = int(ySrc)
        elif attachmentSide == AttachmentSide.WEST:
            xDest   = int(xSrc - LOLLIPOP_LINE_LENGTH)
            yDest   = int(ySrc)
            circleX = int(xSrc - LOLLIPOP_LINE_LENGTH)
            circleY = int(ySrc)
        elif attachmentSide == AttachmentSide.NORTH:
            xDest   = int(xSrc)
            yDest   = int(ySrc - LOLLIPOP_LINE_LENGTH)
            circleX = int(xSrc)
            circleY = int(ySrc - LOLLIPOP_LINE_LENGTH)
        else:  # it is South
            xDest   = int(xSrc)
            yDest   = int(ySrc + LOLLIPOP_LINE_LENGTH)
            circleX = int(xSrc)
            circleY = int(ySrc + LOLLIPOP_LINE_LENGTH)

        return circleX, circleY, xDest, yDest

    def _createTheInterfaceLine(self, position: Position, xDest: int, yDest: int):

        internalDest: InternalPosition = self._pdfCommon.toInternal(position=Position(x=xDest, y=yDest))
        internalSrc:  InternalPosition = self._pdfCommon.toInternal(position=position)
        self._pdf.line(x1=internalSrc.x, y1=internalSrc.y, x2=internalDest.x, y2=internalDest.y)

    def _drawTheTootsiePopCircle(self, attachmentSide: AttachmentSide, circleX: int, circleY: int):

        adjustedX, adjustedY = self._adjustXYForCircle(attachmentSide=attachmentSide, x=circleX, y=circleY)

        internalAdjusted: InternalPosition = self._pdfCommon.toInternal(Position(x=adjustedX, y=adjustedY))
        self._pdf.ellipse(x=internalAdjusted.x, y=internalAdjusted.y,
                          w=LOLLIPOP_CIRCLE_WIDTH, h=LOLLIPOP_CIRCLE_HEIGHT,
                          style=RenderStyle.Draw.value)

    def _drawInterfaceName(self, interfaceName: str, attachmentSide: AttachmentSide, xSrc: int, ySrc: int):

        textWidth = self._pdf.get_string_width(s=interfaceName)
        pixelSize = self._pdf.font_size_pt

        position = self._determineInterfaceNamePosition(xSrc=xSrc, ySrc=ySrc,
                                                        attachmentSide=attachmentSide,
                                                        pixelSize=(pixelSize, pixelSize),
                                                        textSize=(textWidth, pixelSize))

        textPosition: InternalPosition = self._pdfCommon.toInternal(position=position)

        self._pdf.text(x=textPosition.x, y=textPosition.y, txt=interfaceName)

    def _adjustXYForCircle(self, attachmentSide: AttachmentSide, x: int, y: int) -> Tuple[int, int]:

        adjustedX: int = 0
        adjustedY: int = 0

        if attachmentSide == AttachmentSide.EAST:
            adjustedX = x
            adjustedY = y - (LOLLIPOP_CIRCLE_HEIGHT // 2)
        elif attachmentSide == AttachmentSide.WEST:
            adjustedX = x - LOLLIPOP_CIRCLE_WIDTH
            adjustedY = y - (LOLLIPOP_CIRCLE_HEIGHT // 2)
        elif attachmentSide == AttachmentSide.NORTH:
            adjustedX = x - (LOLLIPOP_CIRCLE_WIDTH // 2)
            adjustedY = y - LOLLIPOP_CIRCLE_HEIGHT
        elif attachmentSide == AttachmentSide.SOUTH:
            adjustedX = x - (LOLLIPOP_CIRCLE_WIDTH // 2)
            adjustedY = y

        return adjustedX, adjustedY

    def _determineInterfaceNamePosition(self, xSrc, ySrc,
                                        attachmentSide: AttachmentSide,
                                        pixelSize: Tuple[int, int],
                                        textSize: Tuple[int, int]) -> Position:

        position: Position = Position()
        x:        int      = xSrc
        y:        int      = ySrc

        fWidth, fHeight = pixelSize
        tWidth, tHeight = textSize

        if attachmentSide == AttachmentSide.NORTH:
            y -= (LOLLIPOP_LINE_LENGTH + (LOLLIPOP_CIRCLE_RADIUS * 2) + ADJUST_AWAY_FROM_IMPLEMENTOR)
            x -= (tWidth // 2)
            position.x = x
            position.y = y

        elif attachmentSide == AttachmentSide.SOUTH:
            y += (LOLLIPOP_LINE_LENGTH + LOLLIPOP_CIRCLE_RADIUS + ADJUST_AWAY_FROM_IMPLEMENTOR)
            x -= (tWidth // 2)
            position.x = x
            position.y = y

        elif attachmentSide == AttachmentSide.WEST:
            y = y - (fHeight * 2)
            originalX: int = x
            x = x - LOLLIPOP_LINE_LENGTH - (tWidth // 2)
            while x + tWidth > originalX:
                x -= ADJUST_AWAY_FROM_IMPLEMENTOR
            position.x = x
            position.y = y

        elif attachmentSide == AttachmentSide.EAST:
            y = y - (fHeight * 2)
            x = x + round(LOLLIPOP_LINE_LENGTH * 0.8)
            position.x = x
            position.y = y
        else:
            self.logger.warning(f'Unknown attachment point: {attachmentSide}')
            assert False, 'Unknown attachment point'

        return position




from dataclasses import dataclass
from dataclasses import field

from pyumldiagrams.Common import Common
from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import Position

from pyumldiagrams.Defaults import LEFT_MARGIN
from pyumldiagrams.Defaults import TOP_MARGIN
from pyumldiagrams.Internal import InternalPosition


@dataclass
class Coordinates:
    x: int = 0
    y: int = 0


def createCoordinatesFactory() -> Coordinates:
    return Coordinates()


@dataclass
class Dimensions:
    width:  int = 0
    height: int = 0


def createDimensionsFactory() -> Dimensions:
    return Dimensions()


@dataclass
class PdfShapeDefinition:
    coordinates: Coordinates = field(default_factory=createCoordinatesFactory)
    dimensions:  Dimensions  = field(default_factory=createDimensionsFactory)


class PdfCommon(Common):

    def __init__(self, diagramPadding: DiagramPadding, dpi: int):

        self._diagramPadding: DiagramPadding = diagramPadding
        self._dpi:            int            = dpi

    @classmethod
    def toPdfPoints(cls, pixelNumber: float, dpi: int) -> int:
        """

        points = pixels * 72 / DPI

        Args:
            pixelNumber:  From the display
            dpi:  dots per inch of source display

        Returns:  A pdf point value to use to position on a generated document

        """
        points: int = int((pixelNumber * 72)) // dpi

        return points

    @classmethod
    def convertPosition(cls, pos: Position, dpi: int, verticalGap: int, horizontalGap: int) -> Coordinates:

        x: int = PdfCommon.toPdfPoints(pos.x, dpi) + LEFT_MARGIN + verticalGap
        y: int = PdfCommon.toPdfPoints(pos.y, dpi) + TOP_MARGIN  + horizontalGap

        return Coordinates(x=x, y=y)

    def toInternal(self, position: Position) -> InternalPosition:

        verticalGap:   int = self._diagramPadding.verticalGap
        horizontalGap: int = self._diagramPadding.horizontalGap

        coordinates: Coordinates = PdfCommon.convertPosition(pos=position, dpi=self._dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)

        internalPosition: InternalPosition = InternalPosition(coordinates.x, coordinates.y)

        return internalPosition




from typing import Tuple

from dataclasses import dataclass

from codeallybasic.Position import Position

from codeallyadvanced.ui.Common import Common
from codeallyadvanced.ui.AttachmentSide import AttachmentSide


@dataclass(eq=True)
class AbsolutePosition:
    """
    """
    x: int = 0
    y: int = 0


class CommonAbsolute:
    """
    This class of computations resides here instead of common because it uses the Position
    class from Definitions and not and InternalPosition class which is relative to the
    particular output type (pdf/image)
    """
    X_FUDGE_FACTOR: int = 9
    Y_FUDGE_FACTOR: int = 9

    @classmethod
    def computeAbsoluteLabelPosition(cls, srcPosition: Position, dstPosition: Position, labelPosition: Position) -> Tuple[int, int]:
        """
        Label  positions are relative to the line they are attached to;

        Args:
            srcPosition:
            dstPosition:
            labelPosition:
        """
        xLength: int = abs(srcPosition.x - dstPosition.x)
        yLength: int = abs(srcPosition.y - dstPosition.y)

        if srcPosition.x < dstPosition.x:
            x: int = srcPosition.x + (xLength // 2) + labelPosition.x
            if cls.doXAdjustment(srcPosition=srcPosition, dstPosition=dstPosition) is True:
                x += CommonAbsolute.X_FUDGE_FACTOR
        else:
            x = dstPosition.x + (xLength // 2) + labelPosition.x
            if cls.doXAdjustment(srcPosition=srcPosition, dstPosition=dstPosition) is True:
                x -= CommonAbsolute.X_FUDGE_FACTOR

        if srcPosition.y < dstPosition.y:
            y: int = srcPosition.y + (yLength // 2) + labelPosition.y
        else:
            y = dstPosition.y + (yLength // 2) + labelPosition.y

        y += CommonAbsolute.Y_FUDGE_FACTOR

        return x, y

    @classmethod
    def doXAdjustment(cls, srcPosition: Position, dstPosition: Position) -> bool:

        ans: bool = True

        placement: AttachmentSide = Common.whereIsDestination(sourcePosition=srcPosition, destinationPosition=dstPosition)

        if placement == AttachmentSide.NORTH or placement == AttachmentSide.SOUTH:
            ans = False

        return ans




from typing import List
from typing import NewType
from typing import cast

from dataclasses import dataclass
from dataclasses import field

from enum import Enum

from codeallybasic.Position import Position

from codeallyadvanced.ui.AttachmentSide import AttachmentSide

from pyumldiagrams.Defaults import TOP_MARGIN
from pyumldiagrams.Defaults import LEFT_MARGIN
from pyumldiagrams.Defaults import DEFAULT_HORIZONTAL_GAP
from pyumldiagrams.Defaults import DEFAULT_VERTICAL_GAP
from pyumldiagrams.UnsupportedException import UnsupportedException


ClassName = NewType('ClassName', str)


def createPositionFactory() -> Position:
    return Position()


@dataclass
class DiagramPadding:
    """
    todo::  These should move to the Internal package
    """

    topMargin:  int = TOP_MARGIN
    """
    The diagram's observed top margin.  See `pyumldiagrams.Defaults.TOP_MARGIN`
    """
    leftMargin: int = LEFT_MARGIN
    """
    The diagram's observed left margin.  See `pyumldiagrams.Defaults.LEFT_MARGIN`
    """

    horizontalGap: int = DEFAULT_HORIZONTAL_GAP
    """
    The horizontal gap between UML graphics added to the layout in addition to the gap imposed 
    by the actual graphics positions. See `pyumldiagrams.Defaults.DEFAULT_HORIZONTAL_GAP`
    """
    verticalGap:   int = DEFAULT_VERTICAL_GAP
    """
    The vertical gap between UML graphics added to the layout in addition to the gap imposed 
    by the actual graphics positions.  See `pyumldiagrams.Defaults.DEFAULT_VERTICAL_GAP`
    """


@dataclass
class Size:
    """
    Defines the size of the input UML definitions;
    """
    width:  int = 100
    """
    The width of a shape
    """
    height: int = 100
    """
    The height of the shape
    """


def createSizeFactory() -> Size:
    return Size()


class VisibilityType(Enum):
    """
    Defines the visibility of either methods or fields
    """
    Public    = '+'
    Private   = '-'
    Protected = '#'

    @staticmethod
    def toEnum(strValue: str) -> 'VisibilityType':
        """
        Converts the input string to the line type enum
        Args:
            strValue:   The serialized string representation

        Returns:  The line type enumeration
        """
        canonicalStr: str = strValue.lower().strip(' ')
        if canonicalStr == 'public':
            return VisibilityType.Public
        elif canonicalStr == 'private':
            return VisibilityType.Private
        elif canonicalStr == 'protected':
            return VisibilityType.Protected
        else:
            raise UnsupportedException(f'Do not handle VisibilityType {canonicalStr}')


@dataclass
class BaseDefinition:

    __slots__ = ['name']
    name: str
    """
    The name associated with the definition.
    """


@dataclass
class DefaultValueDefinition(BaseDefinition):
    defaultValue:  str = ''
    """
    A string that describes a parameter default value
    """


@dataclass
class ParameterDefinition(DefaultValueDefinition):
    """
    Defines a single parameter for a method
    """
    parameterType: str = ''
    """
    A string that describes the parameter type
    """


Parameters = NewType('Parameters', List[ParameterDefinition])
"""
Syntactic sugar to define a list of parameters.  
"""


class DisplayMethodParameters(Enum):

    DISPLAY        = 'Display'
    DO_NOT_DISPLAY = 'DoNotDisplay'
    UNSPECIFIED    = 'Unspecified'


def createParametersFactory() -> Parameters:
    return Parameters([])


@dataclass
class MethodDefinition(BaseDefinition):
    """
    Defines a single method in a single UML class
    """
    visibility: VisibilityType = VisibilityType.Public
    """
    Defines the method visibility.  See `VisibilityType`
    """
    returnType: str = ''
    """
    Defines the method return type.
    """
    parameters: Parameters = field(default_factory=createParametersFactory)
    """
    Define the parameters for a particular method
    """


Methods = NewType('Methods', List[MethodDefinition])
"""
Syntactic sugar to define a list of methods.
"""


def createMethodsFactory() -> Methods:
    return Methods([])


@dataclass
class FieldDefinition(DefaultValueDefinition):
    """
    Defines a single instance variable
    """
    fieldType: str = ''
    """
    A string that describes the field type
    """
    visibility: VisibilityType = VisibilityType.Public
    """
    Defines the field visibility.  See `VisibilityType`
    """


Fields = NewType('Fields', List[FieldDefinition])


def createFieldsFactory() -> Fields:
    return Fields([])


@dataclass
class ShapeDefinition(BaseDefinition):
    size:     Size     = field(default_factory=createSizeFactory)
    """
    The size of UML class symbol.  See `Size`
    """
    position: Position = field(default_factory=createPositionFactory)
    """
    The position of the UML class symbol.  See `Position`
    """


@dataclass
class ClassDefinition(ShapeDefinition):
    """ The class definition.  Currently, does not support instance properties.
    """
    methods: Methods   = field(default_factory=createMethodsFactory)
    """
    The list of methods this class implements.  
    """
    fields:  Fields    = field(default_factory=createFieldsFactory)
    """
    The list of instance variables this class defines.
    """
    displayStereotype:  bool = True
    """
    If true display the class stereotype when drawing the class diagram
    """
    displayMethods:     bool = True
    """
    If True display the class methods
    """
    displayFields:      bool = True
    """
    If True display the class instance variables
    """
    displayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.UNSPECIFIED
    """
    If True display the method parameters;  If UNSPECIFIED defer to global
    """
    fileName: str = ''
    """
    The file name where the original source code came from (only if reverse engineered)
    """
    description: str = ''
    """
    Text describing the rationale for this class
    """


ClassDefinitions = NewType('ClassDefinitions', List[ClassDefinition])

Content  = NewType('Content', str)
Contents = NewType('Contents', List[Content])


def createContentsFactory() -> Contents:
    return Contents([])


@dataclass
class NoteDefinition(ShapeDefinition):
    content: Contents = field(default_factory=createContentsFactory)
    """
    The note content
    """


UmlNoteDefinitions = NewType('UmlNoteDefinitions', List[NoteDefinition])


class LineType(Enum):
    """
    The type of UML line you wish to draw.  Currently, bare associations are not supported.
    """
    Inheritance     = 0
    Interface       = 1
    Aggregation     = 3
    Composition     = 7
    Association     = 9
    NoteLink        = 11
    Lollipop        = 13

    @staticmethod
    def toEnum(strValue: str) -> 'LineType':
        """
        Converts the input string to the line type enum
        Args:
            strValue:   The serialized string representation

        Returns:  The line type enumeration
        """
        canonicalStr: str = strValue.lower().strip(' ')
        if canonicalStr == 'aggregation':
            return LineType.Aggregation
        elif canonicalStr == 'composition':
            return LineType.Composition
        elif canonicalStr == 'inheritance':
            return LineType.Inheritance
        elif canonicalStr == 'interface':
            return LineType.Interface
        elif canonicalStr == 'association':
            return LineType.Association
        elif canonicalStr == 'notelink':
            return LineType.NoteLink
        elif canonicalStr == 'lollipop':
            return LineType.Lollipop
        else:
            raise UnsupportedException(f'Do not handle LineType {canonicalStr}')


NamedAssociations: List[LineType] = [LineType.Association, LineType.Aggregation, LineType.Composition]


LinePositions = NewType('LinePositions', list[Position])


@dataclass
class LineDefinition:
    """
    Defines a line between many points;  Index 0 the start of the line;  That last point
    is the end of the line
    """
    linePositions: LinePositions


@dataclass
class UmlLineDefinition(LineDefinition):
    """
    A UML Line definition includes its type
    """
    lineType: LineType
    """
    The UML line type  See `LineType`.
    """
    name: str = ''
    """
    Name of composition, aggregation association links
    """
    cardinalitySource:      str = ''
    cardinalityDestination: str = ''
    """
    Cardinality values used for composition, aggregation, and association links
    """
    namePosition:                   Position = field(default_factory=createPositionFactory)
    sourceCardinalityPosition:      Position = field(default_factory=createPositionFactory)
    destinationCardinalityPosition: Position = field(default_factory=createPositionFactory)


UmlLineDefinitions = NewType('UmlLineDefinitions', List[UmlLineDefinition])
"""
Syntactic sugar to define a list of UML Lines.
"""


@dataclass
class UmlLollipopDefinition:
    name: str = ''
    """
    Interface Name 
    """
    attachmentSide: AttachmentSide = cast(AttachmentSide, None)
    position:       Position       = field(default_factory=createPositionFactory)


UmlLollipopDefinitions = NewType('UmlLollipopDefinitions', List[UmlLollipopDefinition])


class RenderStyle(Enum):
    """
    An enumeration that determines how to draw various UML and other graphical elements
    """
    Draw     = 'D'
    """
    Just draw the outline
    """
    Fill     = 'F'
    """
    Just fill in the area associated with the shape
    """
    DrawFill = 'DF'
    """
    Do both when drawing the UML shape or figure
    """


@dataclass
class RectangleDefinition:
    """
    Defines a rectangle
    """

    renderStyle: RenderStyle = RenderStyle.Draw
    """
    How to draw the rectangle.  See `RenderStyle`
    """
    position:    Position    = field(default_factory=createPositionFactory)
    """
    Where to put the rectangle.  See `Position`
    """
    size:        Size        = field(default_factory=createSizeFactory)
    """
    The rectangle size.  See `Size`
    """


@dataclass
class EllipseDefinition(RectangleDefinition):
    """
    This is just typing syntactical sugar on how to define an Ellipse.
    """
    pass




from typing import cast

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from untangle import Element

from codeallybasic.Common import XML_END_OF_LINE_MARKER

from pyumldiagrams.Definitions import Contents
from pyumldiagrams.Definitions import NoteDefinition
from pyumldiagrams.xmlsupport import XmlConstants
from pyumldiagrams.xmlsupport.ShapeAttributes import ShapeAttributes


class UnTangleNoteDefinition(ShapeAttributes):

    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

    def untangle(self, linkElement: Element) -> NoteDefinition:

        pyutNoteElement: Element = linkElement.PyutNote

        noteDefinition: NoteDefinition = NoteDefinition(name='')    # Notes do not have a name

        noteDefinition.size     = self._shapeSize(graphicElement=linkElement)
        noteDefinition.position = self._shapePosition(graphicElement=linkElement)

        # TODO:  Update when code-ally-basic has common code
        rawContent:   str      = pyutNoteElement[XmlConstants.ATTR_CONTENT_V11]
        cleanContent: str      = rawContent.replace(XML_END_OF_LINE_MARKER, osLineSep)
        lineList:     Contents = cast(Contents, cleanContent.split(osLineSep))      # I am coercive

        noteDefinition.content  = Contents(lineList)

        return noteDefinition




from typing import List
from typing import NewType
from typing import Union

from dataclasses import dataclass

from untangle import Element


@dataclass(eq=True)
class InternalPosition:
    """
    The x and y coordinates are relative to the diagramming method.  For pdf
    documents they are in points.  For images, they are in pixels.  In all cases, the position is
    adjusted for left and right margins plus vertical and horizontal gaps.
    """
    x: int = 0
    y: int = 0


def createInternalPositionFactory() -> InternalPosition:
    return InternalPosition()


@dataclass
class SeparatorPosition(InternalPosition):
    pass


ArrowPoints   = List[InternalPosition]
DiamondPoints = List[InternalPosition]
PolygonPoints = Union[ArrowPoints, DiamondPoints]


Elements = NewType('Elements', List[Element])
